function [ data, g, data0 ] = ...
                    kolmogorovTCP(expectFcn, whichModes, pdrop, tMax, accuracy)
% kolmogorovTCP: solve Kolmogorov eqn for a model of TCP network flow.
%
%   [ data, g, data0 ] = ...
%                   kolmogorovTCP(expectFcn, whichModes, pdrop, tMax, accuracy)
%  
% This function uses the model of TCP network flow rates from the paper
%   Joao P. Hespanha, "Stochastic Hybrid Systems: Application to
%   Communication Networks", pp.387-401, HSCC 2004, LNCS 2993.
%
% Expected outcomes for the model are determined by solving a backwards
%   Kolmogorov equation (derived from the characteristic eqn for the system).
%   This is an HJ PDE with convection, discounting and forcing terms.
%
% The model for transfer file size is a mixed exponential.  In addition
%   to the OFF mode (no transmission), for each component of the exponential
%   there is a pair of modes
%        SS  - slow start.
%        CA  - congestion avoidance.
%   Technically, the OFF mode has no transmission window size w, but
%   we treat the mode as if it had such a continuous dimension.  There
%   is no flow in this dummy dimension.  Note that this treatment
%   requires a bit of attention, because the resulting HJ PDE has no
%   spatial dependence (and thus no CFL restriction).
%
% There is just one continuous variable in each mode (w - window size).
%   Although the model in the paper includes a second variable (s), it
%   turns out that there is no dependence on this variable in any of the
%   dynamics.  Even the rate of transmission completion is independent
%   of s, once you simplify the term.  The metric of interest
%   is the packet transmission rate r = ds/dt = w / RTT.
%
% This function was originally designed as a script file, so most of the
%   options can only be modified in the file.
%
% For example, edit the file to change the grid dimension, boundary conditions,
%   flow field parameters, etc.
%
% Parameters:
%
%   expectFcn    What function would you like to take the expectation of?
%                  'linear'      Compute E[packet rate] (default).
%                  'mean'        Same as 'linear'.
%                  'square'      Compute E[packet rate^2].
%                  'variance'    Same as 'square' (not really the variance).
%   whichModes   Which modes should be included in the expectation?
%                  'all'         All modes except OFF (default).
%                  'SS'          Only slow start modes.
%                  'CA'          Only collision avoidance modes.
%                  an integer    Only modes corresponding to that
%                                  index into the file size array.
%   pdrop        Probability of a dropped packet.  Default = 1e-2.
%   tMax         Time at which to stop integrating.  Default = 8 sec.
%                  Integration may also stop if "convergence" is detected
%                  (see the source for the definition of "convergence").
%   accuracy     Controls the order of approximations.
%                  'low'         Use odeCFL1 and upwindFirstFirst (default).
%                  'medium'      Use odeCFL2 and upwindFirstENO2.
%                  'high'        Use odeCFL3 and upwindFirstENO3.
%                  'veryHigh'    Use odeCFL3 and upwindFirstWENO5.
%
%   data         Implicit surface function at t_max.
%   g            Grid structure on which data was computed.
%   data0        Implicit surface function at t_0.

% Copyright 2004 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Ian Mitchell, 9/13/04.
% Modified to use v1.1 vector level set formulation, 12/11/04.

%---------------------------------------------------------------------------
% You will see many executable lines that are commented out.
%   These are included to show some of the options available; modify
%   the commenting to modify the behavior.
  
%---------------------------------------------------------------------------
% Make sure we can see the kernel m-files.
run('../../addPathToKernel');

%---------------------------------------------------------------------------
% Model parameters
tau_off = 5;                            % Mean off period (s).
RTT = 0.05;                             % Round trip time (s).
n_ack = 1;                              % Number packets acknowledged / packet.
window_0 = 0.693;                       % Initial window size.

% Assume convergence if all data values are within what relative range?
convergedRelativeError = 0.01;

% Probability of each packet being dropped (range 3e-4 to 3e-1).
if(nargin < 3)
  p_drop = 1e-2;
else
  p_drop = pdrop;
end

% Exponential mixture(s).  Vector p_mix should sum to 1.
% The two component mixture from Hespanha.
k_mix = 1024/1500 * [ 3.5; 246 ];       % Mean transfer sizes (packets).
p_mix = [ 0.8887; 0.1113 ];             % Probability of that size transfer.

% The three component mixture from Hespanha.
%k_mix = 1024/1500 * [ 6.0; 400; 10240 ];
%p_mix = [ 0.98; 0.017; 0.003 ];

% How many components are there in the mixed exponential?
numSizes = length(k_mix);

% Mixed exponential distribution function.
%   How do the transition rates depend on k and s?
%   This is the function (DF_s(s) / (1 - F_s(s))) in the paper.
ratio_F = inline('1 ./ k', 'k');

% Indexing constants for each mode.
%   OFF mode is placed last to ensure that its lack of CFL timestep
%   restriction does not cause problems in the temporal integrator.
modeSS  = 2 * (1:numSizes)' - 1;
modeCA  = 2 * (1:numSizes)';
modeOFF = 2 * numSizes + 1;

% How many different modes in the model?
numModes = 2 * numSizes + 1;

%---------------------------------------------------------------------------
% End time for integration.
if(nargin < 4)
  tMax = 8.0;
end

% Integration parameters.
plotSteps = 16;              % How many intermediate plots to produce?
t0 = 0;                      % Start time.
singleStep = 0;              % Plot at each timestep (overrides tPlot).

% Period at which intermediate plots should be produced.
tPlot = (tMax - t0) / (plotSteps - 1);

% How close (relative) do we need to get to tMax to be considered finished?
small = 100 * eps;

%---------------------------------------------------------------------------
% We are not viewing level sets, so choose level to disable warnings.
level = [];

% Pause after each plot?
pauseAfterPlot = 0;

% Delete previous plot before showing next?
deleteLastPlot = 0;

% Plot in separate subplots (set deleteLastPlot = 0 in this case)?
useSubplots = 1;

%---------------------------------------------------------------------------
% What kind of display?
displayType = 'plot';

styleName = { 'LineStyle', 'Color' };
style = cell(numModes, 1);
colors = { 'blue', 'red', 'black', 'magenta', 'cyan', 'green', 'yellow' };
for i = 1 : numSizes
  style{modeSS(i)}  = {  '-', colors{i} };
  style{modeCA(i)}  = { '--', colors{i} };
end

style{modeOFF} = { ':', 'magenta' };

%---------------------------------------------------------------------------
if(nargin < 5)
  accuracy = 'low';
end

% Choose approximations at appropriate level of accuracy.
switch(accuracy)
 case 'low'
  derivFunc = @upwindFirstFirst;
  integratorFunc = @odeCFL1;
 case 'medium'
  derivFunc = @upwindFirstENO2;
  integratorFunc = @odeCFL2;
 case 'high'
  derivFunc = @upwindFirstENO3;
  integratorFunc = @odeCFL3;
 case 'veryHigh'
  derivFunc = @upwindFirstWENO5;
  integratorFunc = @odeCFL3;
 otherwise
  error('Unknown accuracy level %s', accuracy);
end

%---------------------------------------------------------------------------
% Set up time approximation scheme.
integratorOptions = odeCFLset('factorCFL', 0.75, 'stats', 'on');

if(singleStep)
  integratorOptions = odeCFLset(integratorOptions, 'singleStep', 'on');
end

%---------------------------------------------------------------------------
% Create the grid.
%   Grid resolution and expanse is difficult to determine.
%   Experimental evidence seems to suggest
%     1) Larger g.max leads to larger stationary solutions.
%     2) Larger g.N leads to smaller stationary solutions.
%   However, the differences for g.N ~ 200 and g.max ~ 100 were
%     less than 1% of the final stationary value.
g.dim = 1;
g.N = 200;
g.max = 100;
g.min = g.max / g.N(1);
g.bdry = @addGhostMixedDirichletExtrapolate;
g.bdryData.order = 1;
g = processGrid(g);

% For setting model parameters, create local names.
% Which grid dimension is associated with window size?
indexW = 1;

% Shorthand name for window size of each node.
gridW = g.xs{indexW};

% For terms representing dropped packets, we need to build a grid 
%   representing the state after the window size has been reset to 
%   half its previous size.
xsHalf = g.xs;
xsHalf{indexW} = 0.5 * g.xs{indexW};

%---------------------------------------------------------------------------
% Switch rates.  There are three types of switches.

% Initiating a transmission: from OFF mode to one of the SS_i.
lambda_s = p_mix / tau_off;

% Dropped packet: from SS_i or CA_i to CA_i.
lambda_d = p_drop * gridW / RTT;

% Finished Transmission: from SS_i or CA_i to OFF.
lambda_o = cell(numSizes, 1);
for i = 1 : numSizes
  lambda_o{i} = gridW * ratio_F(k_mix(i)) / RTT;
end

%---------------------------------------------------------------------------
% Mode dynamics: all modes.

% All modes will consist of a sum of other terms.
schemeFunc = @termSum;

% The modes will have different data driving their dynamics.
schemeData = cell(numModes, 1);

%---------------------------------------------------------------------------
% Mode dynamics: mode SS.

% There are three classes of terms: convective, discount and forcing.
%   The first term is independent of transmission file size.

% Convective terms (continuous dynamics).
convectDataSS.grid = g;
convectDataSS.derivFunc = derivFunc;
convectDataSS.velocity = { -log(2) * gridW / (n_ack * RTT) };

% The remaining terms depend on file size through lambda_o
for size = 1 : numSizes

  % Discount terms -- the HJ component for switches out of this mode 
  %   that depends on the value function in this mode.
  discountDataSS.grid = g;
  discountDataSS.lambda = lambda_o{size} + lambda_d;

  % Forcing terms -- the HJ component for switches out of this mode
  %   that depend on other elements of the vector level set.
  %   We effectively have a time-dependent forcing function.
  forceDataSS.grid = g;
  forceDataSS.xsHalf = xsHalf;
  forceDataSS.mapFactor = 1.0;
  forceDataSS.lambda_d = lambda_d;
  forceDataSS.lambda_o = lambda_o{size};
  forceDataSS.forcing = @forcingFuncSS;

  % Note that we need to give the indices of the OFF mode and the matching
  %   CA mode in the vector level set relative to this mode's index.
  % The matching CA mode is always the very next vector level set entry.
  forceDataSS.modeCA = 2;
  forceDataSS.modeOFF = modeOFF - modeSS(size) + 1;

  % We will need access to the entire vector level set in order
  %   to compute the dependencies for switches out of this mode.
  forceDataSS.passVLS = 1;

  % Sum the three terms together.
  schemeData{modeSS(size)}.innerFunc = ...
                              { @termConvection; @termDiscount; @termForcing };
  schemeData{modeSS(size)}.innerData = ...
                                { convectDataSS; discountDataSS; forceDataSS };

  % We also need the grid at this level of the schemeData cell/structure 
  %   for reshaping the vector level inside termForcing.
  schemeData{modeSS(size)}.grid = g;

end

%---------------------------------------------------------------------------
% Mode dynamics: mode CA.

% There are three classes of terms: convective, discount and forcing.
%   The first term is independent of transmission file size.

% Convective terms (continuous dynamics).
convectDataCA.grid = g;
convectDataCA.derivFunc = derivFunc;
convectDataCA.velocity = { -1 / (n_ack * RTT) };

% The remaining terms depend on file size through lambda_o
for size = 1 : numSizes

  % Discount terms -- the HJ component for switches out of this mode 
  %   that depends on the value function in this mode.
  discountDataCA.grid = g;
  discountDataCA.lambda = lambda_o{size} + lambda_d;

  % Forcing terms -- the HJ component for switches out of this mode
  %   that depend on other elements of the vector level set
  %   and also one term dependent on a mapped version of this element.
  %   We effectively have a time-dependent forcing function.
  forceDataCA.grid = g;
  forceDataCA.xsHalf = xsHalf;
  forceDataCA.mapFactor = 1.0;
  forceDataCA.lambda_d = lambda_d;
  forceDataCA.lambda_o = lambda_o{size};
  forceDataCA.forcing = @forcingFuncCA;

  % Note that we need to give the indices of the OFF mode
  %   in the vector level set relative to this mode's index.
  forceDataCA.modeOFF = modeOFF - modeCA(size) + 1;

  % We will need access to the entire vector level set in order
  %   to compute the dependencies for switches out of this mode.
  forceDataCA.passVLS = 1;

  % Sum the three terms together.
  schemeData{modeCA(size)}.innerFunc = ...
                              { @termConvection; @termDiscount; @termForcing };
  schemeData{modeCA(size)}.innerData = ...
                                { convectDataCA; discountDataCA; forceDataCA };

  % We also need the grid at this level of the schemeData cell/structure 
  %   for reshaping the vector level inside termForcing.
  schemeData{modeCA(size)}.grid = g;

end

%---------------------------------------------------------------------------
% Mode dynamics: mode OFF.

% This mode has no convective term, but does have discount and forcing terms.

% When a transmission first starts, it jumps to a specific window size.
state_0 = window_0;

% Discount terms -- the HJ component for switches out of this mode 
%   that depends on the value function in this mode.
discountDataOFF.grid = g;
discountDataOFF.lambda = sum(lambda_s);

% Forcing terms -- the HJ component for switches out of this mode
%   that depend on other elements of the vector level set.
%   We effectively have a time-dependent forcing function.
forceDataOFF.grid = g;
forceDataOFF.state_0 = state_0;
forceDataOFF.lambda_s = lambda_s;
forceDataOFF.forcing = @forcingFuncOFF;

% Note that we need to give the indices of the SS modes
%   in the vector level set relative to this mode's index.
forceDataOFF.modeSS = modeSS - modeOFF + numModes + 1;

% We will need access to the entire vector level set in order
%   to compute the dependencies for switches out of this mode.
forceDataOFF.passVLS = 1;

% Sum the two terms together.
schemeData{modeOFF}.innerFunc = { @termDiscount; @termForcing };
schemeData{modeOFF}.innerData = { discountDataOFF; forceDataOFF };

% We also need the grid at this level of the schemeData cell/structure 
%   for reshaping the vector level inside termForcing.
schemeData{modeOFF}.grid = g;

%---------------------------------------------------------------------------
% What type of expectation do we want to take?
if(nargin < 1)
  expectFcn = 'linear';
end

switch(expectFcn)
 case { 'linear', 'mean' }
  data0 = gridW / RTT;

 case { 'square', 'variance' }
  data0 = (gridW / RTT).^2;

 otherwise
  error('Unknown expectFcn parameter: %s', expectFcn);
end

%---------------------------------------------------------------------------
% For what modes do we want this expectation to apply?
if(nargin < 2)
  whichModes = 'all';
end

% Cell vector of solutions.
data = cell(numModes, 1);

% Off mode is never included in expectation.
data{modeOFF} = zeros(g.shape);

if(ischar(whichModes))
  switch(whichModes)
   case 'all'
    % All active modes included.
    [ data{[ modeSS; modeCA ]} ] = deal(data0);

   case 'SS'
    % Only slow start modes included.
    [ data{modeSS} ] = deal(data0);
    [ data{modeCA} ] = deal(zeros(g.shape));

   case 'CA'
    % Only congestion avoidance modes included.
    [ data{modeCA} ] = deal(data0);
    [ data{modeSS} ] = deal(zeros(g.shape));

   otherwise
    error('Unknown whichModes parameter: %s', whichModes);

  end
else
  if((whichModes < 1) | (whichModes > numSizes))
    error('Valid whichModes indices are 1 to %d inclusive.', numSizes);
  elseif(fix(whichModes) ~= whichModes)
    error('Parameter whichModes must be an integer if it is an index');
  else
    % Do not include any mode.
    [ data{:} ] = deal(zeros(g.shape));

    % Put back the SS and CA modes corresponding to this file size index.
    [ data{[ modeSS(whichModes); modeCA(whichModes) ]} ] = deal(data0);
  end
end

%---------------------------------------------------------------------------
% Initialize Display
h = cell(numModes, 1);
f = figure;

% Set up subplot parameters if necessary.
if(useSubplots)
  rows = ceil(sqrt(plotSteps));
  cols = ceil(plotSteps / rows);
  plotNum = 1;
  subplot(rows, cols, plotNum);
end

for i = 1 : numModes
  h{i} = visualizeLevelSet(g, data{i}, displayType, level, 't = 0');
  hold on;
  setProperty(h{i}, styleName, style{i});
end

%---------------------------------------------------------------------------
% Loop until tMax (subject to a little roundoff).
tNow = t0;
startTime = cputime;
y0 = cell(numModes, 1);
while(tMax - tNow > small * tMax)

  % Reshape data array into column vector for ode solver call.
  for i = 1 : numModes
    y0{i} = data{i}(:);
  end

  % How far to step?
  tSpan = [ tNow, min(tMax, tNow + tPlot) ];
  
  % Take a timestep.
  [ t, y, schemeData ] = feval(integratorFunc, schemeFunc, tSpan, y0,...
                               integratorOptions, schemeData);
  tNow = t(end);


  % Get back the correctly shaped data array
  for i = 1 : numModes
    data{i} = reshape(y{i}, g.shape);
  end

  if(pauseAfterPlot)
    % Wait for last plot to be digested.
    pause;
  end

  % Create new visualizations.
  figure(f);
  if(useSubplots)
    plotNum = plotNum + 1;
    subplot(rows, cols, plotNum);
  end

  for i = 1 : numModes
    if(deleteLastPlot)
      delete(h{i});
    end
    h{i} = visualizeLevelSet(g, data{i}, displayType, level, ...
                             [ 't = ' num2str(tNow) ]);
    hold on;
    setProperty(h{i}, styleName, style{i});
  end

  % Check to see if we are effectively converged.
  %   Convergence occurs when all the nodes in all the modes reach
  %   essentially a constant value.  In other words, all initial transients 
  %   have disappeared from the expected value.
  maxVal = -inf;
  minVal = +inf;
  for i = 1 : numModes
    maxVal = max(max(data{i}(:)), maxVal);
    minVal = min(min(data{i}(:)), minVal);
  end
  if(maxVal - minVal < convergedRelativeError * maxVal)
    break;
  end

end

endTime = cputime;
fprintf('Total execution time %g seconds', endTime - startTime);



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function forcing = forcingFuncSS(t, data, schemeData)
% forcingFuncSS: compute the forcing term for the SS mode in TCP model.
%
%  forcing = forcingFuncSS(t, data, schemeData)
%
% This function computes the forcing term for the SS mode in the 
%   stochastic hybrid system model of TCP flow proposed by Hespanha.
%
% The forcing terms include switches to other modes:
%        to OFF at rate lambda_o,
%        to CA at rate lambda_d.
%
% This function implements the scalarGridFunc prototype.
%
% Because this function is designed to work on a vector level set
%   description of the state of a hybrid system, the data and schemeData
%   parameters are cell vectors.  The first element corresponds to the
%   current SS mode, while the remaining elements correspond to other modes.
%
% Parameters:
%   t            Current time.
%   data         Vector level set function (a cell vector).
%   schemeData   Cell vector of structures (see below).
%
%   forcing      Array containing the forcing function value at each grid node.
%
% schemeData{1} is a structure containing data specific to this type of 
%   term approximation.  For this function it contains the field(s)
%
%   .grid	 Grid structure (see processGrid.m for details).
%   .xsHalf      Cell vector of arrays storing the grid node locations after
%                  the window size has been divided by two.
%   .mapFactor   Multiplier caused by window size division.
%   .modeCA      Index of CA element of vector level set (scalar).
%   .modeOFF     Index of OFF element of vector level set (scalar).
%   .lambda_d    Rate at which packets are dropped (array same size as data).
%   .lambda_o    Rate at which transmissions finish (array same size as data).
%
% schemeData{1} may contain other fields.

  % Note that we are receiving a cell vector of data and schemeData.
  %   The first element of these cell vectors is the active one.
  % We don't actually need the other elements of schemeData, but we
  %   do need the other elements of data, so we have no choice.

  checkStructureFields(schemeData{1}, 'grid', 'xsHalf', 'mapFactor', ...
                                      'modeCA', 'modeOFF', ...
                                      'lambda_d', 'lambda_o');
  
  %---------------------------------------------------------------------------
  % Switch to CA mode.
  
  % During switch to CA mode, the window size is reset to half its former size.
  phi_CA = data{schemeData{1}.modeCA};
  phi_CA_over2 = interp1(schemeData{1}.grid.xs{:}, ...
                         phi_CA, schemeData{1}.xsHalf{:});

  % The interpolation will fail for the smallest values of w/2, since those
  %   values require extrapolation.
  % As a first order approximation, just copy the smallest known value.
  phi_CA_over2(1) = phi_CA(1);

  forcing = -schemeData{1}.lambda_d .* phi_CA_over2 .* schemeData{1}.mapFactor;

  %---------------------------------------------------------------------------
  % Switch to OFF mode.
  %   This switch involves a reset to window size zero.
  %   The element of the vector level set corresponding to OFF mode
  %   should be constant, so we can sample the value from any of its nodes.
  phi_OFF = data{schemeData{1}.modeOFF}(1);
  forcing = forcing - schemeData{1}.lambda_o * phi_OFF;  



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function forcing = forcingFuncCA(t, data, schemeData)
% forcingFuncCA: compute the forcing term for the CA mode in TCP model.
%
%  forcing = forcingFuncCA(t, data, schemeData)
%
% This function computes the forcing term for the CA mode in the 
%   stochastic hybrid system model of TCP flow proposed by Hespanha.
%
% The forcing terms include switches to other modes:
%        to OFF at rate lambda_o,
%        to CA at rate lambda_d (see below).
%
% Note that there is a discrete switch from CA back to CA at rate lambda_d,
%   but with a reset whereby the window size is halved.  Although it looks
%   like a discount factor, this switch cannot be directly handled by
%   a discount factor term because it involves a reset; therefore, it
%   was placed into the forcing function term.
%
% Furthermore, this term is somewhat suspect because it may violate the
%   monotonicity requirement placed on terms involving the level set
%   function value directly.  However, it seems to work it practice.
%   Presumeably its effect is stabilized by the matching discount factor
%   that models the outgoing switch (and satisfies the monotonicity property).
%
% This function implements the scalarGridFunc prototype.
%
% Because this function is designed to work on a vector level set
%   description of the state of a hybrid system, the data and schemeData
%   parameters are cell vectors.  The first element corresponds to the
%   current SS mode, while the remaining elements correspond to other modes.
%
% Parameters:
%   t            Current time.
%   data         Vector level set function (a cell vector).
%   schemeData   Cell vector of structures (see below).
%
%   forcing      Array containing the forcing function value at each grid node.
%
% schemeData{1} is a structure containing data specific to this type of 
%   term approximation.  For this function it contains the field(s)
%
%   .grid	 Grid structure (see processGrid.m for details).
%   .xsHalf      Cell vector of arrays storing the grid node locations after
%                  the window size has been divided by two.
%   .mapFactor   Multiplier caused by window size division.
%   .modeOFF     Index of OFF element of vector level set (scalar).
%   .lambda_d    Rate at which packets are dropped (array same size as data).
%   .lambda_o    Rate at which transmissions finish (array same size as data).
%
% schemeData{1} may contain other fields.

  % Note that we are receiving a cell vector of data and schemeData.
  %   The first element of these cell vectors is the active one.
  % We don't actually need the other elements of schemeData, but we
  %   do need the other elements of data, so we have no choice.

  checkStructureFields(schemeData{1}, 'grid', 'xsHalf', 'mapFactor',...
                                      'modeOFF', 'lambda_d', 'lambda_o');
  
  %---------------------------------------------------------------------------
  % Switch to CA mode.
  
  % During switch to CA mode, the window size is reset to half its former size.
  phi_CA = data{1};
  phi_CA_over2 = interp1(schemeData{1}.grid.xs{:}, ...
                         phi_CA, schemeData{1}.xsHalf{:});

  % The interpolation will fail for the smallest values of w/2, since those
  %   values require extrapolation.
  % As a first order approximation, just copy the smallest known value.
  phi_CA_over2(1) = phi_CA(1);

  forcing = -schemeData{1}.lambda_d .* phi_CA_over2 .* schemeData{1}.mapFactor;

  %---------------------------------------------------------------------------
  % Switch to OFF mode.
  %   This switch involves a reset to window size zero.
  %   The element of the vector level set corresponding to OFF mode
  %   should be constant, so we can sample the value from any of its nodes.
  phi_OFF = data{schemeData{1}.modeOFF}(1);
  forcing = forcing - schemeData{1}.lambda_o * phi_OFF;  



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function forcing = forcingFuncOFF(t, data, schemeData)
% forcingFuncOFF: compute the forcing term for the OFF mode in TCP model.
%
%  forcing = forcingFuncOFF(t, data, schemeData)
%
% This function computes the forcing term for the OFF mode in the 
%   stochastic hybrid system model of TCP flow proposed by Hespanha.
%
% The forcing terms include switches to other modes:
%        to SS at rate lambda_s (see below).
%
% Note that the switch to mode SS involves a reset of the window size
%   to window_0.  
%
% This function implements the scalarGridFunc prototype.
%
% Because this function is designed to work on a vector level set
%   description of the state of a hybrid system, the data and schemeData
%   parameters are cell vectors.  The first element corresponds to the
%   OFF mode, while the remaining elements correspond to other modes.
%
% Parameters:
%   t            Current time.
%   data         Vector level set function (a cell vector).
%   schemeData   Cell vector of structures (see below).
%
%   forcing      Array containing the forcing function value at each grid node.
%
% schemeData{1} is a structure containing data specific to this type of 
%   term approximation.  For this function it contains the field(s)
%
%   .grid	 Grid structure (see processGrid.m for details).
%   .state_0     Initial state at start of a transmission.
%   .modeSS      Vector of indices of SS_i modes for different transfer sizes.
%   .lambda_s    Vector of rates at which transmissions begin.
%
% schemeData{1} may contain other fields.

  % Note that we are receiving a cell vector of data and schemeData.
  %   The first element of these cell vectors is the active one.

  checkStructureFields(schemeData{1}, 'grid', 'state_0', 'modeSS', 'lambda_s');

  %---------------------------------------------------------------------------
  % Initialize forcing function.
  forcing = zeros(schemeData{1}.grid.shape);
  
  %---------------------------------------------------------------------------
  % Deal with each of the SS_i modes.
  for size = 1 : length(schemeData{1}.modeSS)
    phi_SS = data{schemeData{1}.modeSS(size)};
    phi_SS_state_0 = interp1(schemeData{1}.grid.xs{:}, ...
                             phi_SS, schemeData{1}.state_0);

    forcing = forcing - schemeData{1}.lambda_s(size) .* phi_SS_state_0;
  end



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function setProperty(handle, propertyNames, propertyValues)
% setProperty: set handle graphics property settings.
%
%   setProperty(handle, propertyNames, propertyValues)
%
% For a given handle graphics object, sets one or more of that object's
%   graphics properties to particular values.
%
% Parameters:
%   handle          Graphics object handle(s).
%   propertyNames   Cell vector of names of the properties to be set.
%   propretyValues  Cell vector of values of the corresponding properties.

for i = 1 : length(propertyNames)
  for j = 1 : length(handle)
    set(handle(j), propertyNames{i}, propertyValues{i});
  end
end
drawnow;
