function [ integrals, trace, ts ] = simulateSeparate(maxT, showWhat)
% simulateSeparate: simulate an individual trace of the TCP hybrid model.
%
%   [ integrals, trace, ts ] = simulateSeparate(maxT, showWhat)
%
% This routine runs an individual trajectory for the model of TCP network 
%   flow rates from the paper:
%     Joao P. Hespanha, "Stochastic Hybrid Systems: Application to
%     Communication Networks", pp.387-401, HSCC 2004, LNCS 2993.
%
% This is a separated model in the sense that we have separate modes
%   for each component of the file size distribution.
%
% The components of the trace are (in order):
%         window size, 
%         mode (1 = SS, 2 = CA, 3 = OFF),
%         current file transfer size (0 for mode OFF).
%
% The components of the integral vector are (in order):
%         integral of window size over all modes,
%         integral of (window size)^2 over all modes,
%         integral of window size over all SS modes, 
%         integral of (window size)^2 over all SS modes, 
%         integral of window size over all CA modes,
%         integral of (window size)^2 over all CA modes,
%
%   The remaining components of the integral vector come in groups of two,
%     one group for each component of the file transfer size distribution.
%         integral of window size over transfer size i modes,
%         integral of (window size)^2 over transfer size i modes,
%
% Parameters:
%
%   maxT         How long should the simulation run?
%   showPlot     String specifying what plots to display
%                  'none'  - No plotting during simulation (fastest).
%                  'trace' - Plot trace results only.
%                  'ints'  - Plot integrals results only (default).
%                  'all'   - Plot all available plots.
%
%   integrals    Integrals over the trace.
%   trace        The trace of the simulation.
%   ts           The vector of times corresponding to trace rows.


% Copyright 2004 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Ian Mitchell, 9/13/04

  %---------------------------------------------------------------------------
  % Model parameters
  tau_off = 5;                          % Mean off period (s).
  RTT = 0.05;                           % Round trip time (s).
  n_ack = 1;                            % Number packets acknowledged / packet.
  window_0 = 0.693;                     % Initial window size.
  
  % Probability of each packet being dropped (range 3e-4 to 3e-1).
  p_drop = 1e-2;

  % Exponential mixture(s).  Vector p_mix should sum to 1.
  k_mix = 2/3 * [ 3.5; 246 ];                   % Mean transfer sizes (kb).
  p_mix = [ 0.8887; 0.1113 ];             % Probability of that size transfer.
  
  %k_mix = [ 6.0; 400; 10240 ];
  %p_mix = [ 0.98; 0.017; 0.003 ];

  % How many different file sizes
  numSizes = length(k_mix);

  % Cumulative probability sum for choosing file size.
  cumulative_p_mix = cumsum(p_mix);
  
  % Mixed exponential distribution function.
  %   How do the transition rates depend on k and s?
  %   This is the function (DF_s(s) / (1 - F_s(s))) in the paper.
  ratio_F = inline('1 / k', 'k');
  
  % How many different modes in the model?
  numModes = 3;
  
  % Constants for each mode.
  %   OFF mode is placed last to ensure that its lack of CFL timestep
  %   restriction does not cause problems in the temporal integrator.
  modeSS  = 1;
  modeCA  = 2;
  modeOFF = 3;
  
  %---------------------------------------------------------------------------
  % Simulation parameters
  deltaT = RTT / 50;

  % How many steps before a plot is displayed?
  plotPeriod = maxT / 10 / deltaT;
 
  if(nargin < 1)
    maxT = 100;
  end

  if(nargin < 2)
    showWhat = 'integrals';
  end

  if(strcmp(showWhat, 'none'))
    showPlot = 0;
  else
    showPlot = 1;
  end
  figTrace = [];
  figIntegrals = [];

  % Do we actually need all the memory to store the complete trace?
  if(strcmp(showWhat, 'trace') | strcmp(showWhat, 'all') | (nargout > 1))
    keepTrace = 1;
  else
    keepTrace = 0;
  end

  %---------------------------------------------------------------------------
  % Where will we place results (ie what indices in the trace?)
  indexW = 1;
  indexMode = 2;
  indexSize = 3;

  % A separate set of indices for the integrals.
  indexIntW = 1;
  indexIntW2 = 2;
  indexIntWSS = 3;
  indexIntW2SS = 4;
  indexIntWCA = 5;
  indexIntW2CA = 6;

  numIndices = 6;

  % Indices for the various transfer sizes.
  indexIntSizeW = numIndices + 2 * (1 : numSizes)' - 1;
  indexIntSizeW2 = indexIntSizeW + 1;

  %---------------------------------------------------------------------------
  % Initialize in the OFF mode.
  trace0 = [ 0; modeOFF; 0 ];

  % All integrals start at 0.
  integrals = zeros(numIndices + 2 * numSizes, 1);

  nowT = 0;
  step = 1;

  if(keepTrace)
    trace = zeros(ceil(maxT / deltaT), length(trace0));
    trace(step, :) = trace0';
    ts = zeros(ceil(maxT / deltaT), 1);
    ts(step) = 0;
  end

  %---------------------------------------------------------------------------
  % Loop until finished
  nowMode = trace0(indexMode);
  nowW = trace0(indexW);
  nowSize = trace0(indexSize);       % Actually an index into the size array.
  while(nowT < maxT)

    % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    % Get a random number.
    prob = rand(1);

    % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    % Determine whether a transition took place.
    transition = 0;

    if(nowMode == modeOFF)
      lambda_s = 1 / tau_off;
      if(prob < lambda_s * deltaT)
        transition = modeSS;
      end
    else
      lambda_d = p_drop * nowW / RTT;
      lambda_o = nowW * ratio_F(k_mix(nowSize)) / RTT;

      if(prob < lambda_d * deltaT)
        transition = modeCA;
      elseif(prob - lambda_d * deltaT < lambda_o * deltaT)
        transition = modeOFF;
      end
    end

    % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    % Update the state based on the type of transition.
    switch(transition)

     case modeOFF
      nowMode = modeOFF;
      nowW = 0;
      nowSize = 0;

     case modeSS
      nowMode = modeSS;
      nowW = window_0;
      % Need to choose file transfer size.
      prob = rand(1);
      nowSize = min(find(prob <= cumulative_p_mix));

     case modeCA
      nowMode = modeCA;
      nowW = nowW / 2;
      nowSize = nowSize;

     otherwise
      % No transition
      nowMode = nowMode;
      nowSize = nowSize;

      % Window size depends on current mode.
      switch(nowMode)
       case modeOFF
        nowW = nowW;
       case modeSS
        nowW = nowW + deltaT * log(2) * nowW / (n_ack * RTT);
       case modeCA
        nowW = nowW + deltaT / (n_ack * RTT);
       otherwise
        error('Unknown mode %d detected', nowMode);
      end

    end

    % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    % Update trace.
    nowT = nowT + deltaT;
    step = step + 1;

    if(keepTrace)
      ts(step) = nowT;
      trace(step, indexW) = nowW;
      trace(step, indexMode) = nowMode;
      if(nowSize == 0)
        trace(step, indexSize) = 0;
      else
        trace(step, indexSize) = k_mix(nowSize);
      end   
    end
    
    % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    % Update generic integrals.

    nowRate = nowW / RTT;

    integrals(indexIntW) = integrals(indexIntW) + deltaT * nowRate;
    integrals(indexIntWSS) = integrals(indexIntWSS) ...
                               + (nowMode == modeSS) * deltaT * nowRate;
    integrals(indexIntWCA) = integrals(indexIntWCA) ...
                               + (nowMode == modeCA) * deltaT * nowRate;

    integrals(indexIntW2) = integrals(indexIntW2) + deltaT * nowRate^2;
    integrals(indexIntW2SS) = integrals(indexIntW2SS) ...
                               + (nowMode == modeSS) * deltaT * nowRate^2;
    integrals(indexIntW2CA) = integrals(indexIntW2CA) ...
                               + (nowMode == modeCA) * deltaT * nowRate^2;

    % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    % Update file size specific integrals.
    %   Only the terms for the current file size will change.
    if(nowSize ~= 0)
      integrals(indexIntSizeW(nowSize)) = ...
        integrals(indexIntSizeW(nowSize)) + deltaT * nowRate;
      integrals(indexIntSizeW2(nowSize)) = ...
        integrals(indexIntSizeW2(nowSize)) + deltaT * nowRate^2;
    end

    % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    % Plot results (if necessary).
    if(showPlot && (rem(step, plotPeriod) == 0))

      switch(showWhat)
       case 'trace'
         figTrace = showTrace(trace, ts, step, plotPeriod, figTrace);

       case 'ints'
         figIntegrals = showIntegrals(integrals, nowT, figIntegrals);

       case 'all'
         figTrace = showTrace(trace, ts, step, plotPeriod, figTrace);
         figIntegrals = showIntegrals(integrals, nowT, figIntegrals);

       otherwise
         error('Unknown showWhat parameter string: %s', showWhat);

      end

      drawnow;
    end

  end



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function fig = showTrace(trace, ts, step, plotPeriod, fig)
% showTrace: plots the latest results from the evolving trace.

  % Create a figure if necessary.
  if(isempty(fig))
    fig = figure;
  else
    figure(fig);
  end

  % Only plot the most recent results.
  if(step - plotPeriod < 1)
    range = 1 : step;
  else
    range = step-plotPeriod : step;
  end
  
  subplot(3,1,1)
  plot(ts(range), trace(range, 1), 'b-');
  title('window size');
  hold on;
  
  subplot(3,1,2)
  plot(ts(range), trace(range, 2), 'b-');
  title('mode');
  hold on;

  subplot(3,1,3)
  semilogy(ts(range), trace(range, 3), 'b*');
  title('transfer file size');
  hold on;



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function fig = showIntegrals(integrals, nowT, fig)
% showIntegrals: plots the latest integrals from the evolving trace.

  % Show the variances as well?
  showVariance = 1;

  % What about some line styles?
  styles = { 'b*', 'r+', 'ko', 'ms', 'cv', 'g>' };

  % Create a figure if necessary.
  if(isempty(fig))
    fig = figure;
  else
    figure(fig);
  end

  % How many integrals are there?
  numInts = length(integrals) / 2;

  % Add the latest results
  if(showVariance)
    subplot(2, 1, 2)
    for i = 1 : numInts
      plot(nowT, integrals(2 * i) / nowT, styles{i});
      hold on;
    end
    title('Variances');

    subplot(2, 1, 1)
    title('Means');
  end

  for i = 1 : numInts
    plot(nowT, integrals(2 * i - 1) / nowT, styles{i});
    hold on;
  end
